/*
 * DPVS is a software load balancer (Virtual Server) based on DPDK.
 *
 * Copyright (C) 2017 iQIYI (www.iqiyi.com).
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#ifndef __DPVS_DEST_H__
#define __DPVS_DEST_H__

#include "conf/dest.h"
#include "ipvs/service.h"

#include "conf/common.h"
#include "list.h"
#include "dpdk.h"

struct dp_vs_dest {
    struct list_head    n_list;     /* for the dests in the service */

    int                 af;         /* address family */
    /*
     * normally, addr/port is for Real Server,
     * but for SNAT, addr/port is the "to-source"
     * (the target source ip/port translated to).
     */
    union inet_addr     addr;       /* IP address of the server */
    uint16_t            port;       /* port number of the server */

    volatile unsigned   flags;      /* dest status flags */
    rte_atomic16_t      conn_flags; /* flags to copy to conn */
    rte_atomic16_t      weight;     /* server weight */

    rte_atomic32_t      refcnt;     /* reference counter */
    struct dp_vs_stats  stats;      /* Use per-cpu statistics for destination server */

    enum dpvs_fwd_mode  fwdmode;

    /* connection counters and thresholds */
    rte_atomic32_t      actconns;   /* active connections */
    rte_atomic32_t      inactconns; /* inactive connections */
    rte_atomic32_t      persistconns;   /* persistent connections */
    uint32_t            max_conn;   /* upper threshold */
    uint32_t            min_conn;   /* lower threshold */

    /* for virtual service */
    uint16_t            proto;      /* which protocol (TCP/UDP) */
    uint16_t            vport;      /* virtual port number */
    uint32_t            vfwmark;    /* firewall mark of service */
    struct dp_vs_service *svc;      /* service it belongs to */
    union inet_addr     vaddr;      /* virtual IP address */
    unsigned            conn_timeout; /* conn timeout copied from svc*/
    unsigned            limit_proportion; /* limit copied from svc*/
} __rte_cache_aligned;

static inline bool
dp_vs_dest_is_avail(struct dp_vs_dest *dest)
{
    return (dest->flags & DPVS_DEST_F_AVAILABLE) ? true : false;
}

static inline bool
dp_vs_dest_is_overload(struct dp_vs_dest *dest)
{
    return (dest->flags & DPVS_DEST_F_OVERLOAD) ? true : false;
}

static inline int16_t
dp_vs_dest_get_weight(struct dp_vs_dest *dest)
{
    return rte_atomic16_read(&dest->weight);
}

static inline bool
dp_vs_dest_is_valid(struct dp_vs_dest *dest)
{
    return (dest
            && dp_vs_dest_is_avail(dest)
            && !dp_vs_dest_is_overload(dest)
            && dp_vs_dest_get_weight(dest) > 0) ? true : false;
}

int dp_vs_new_dest(struct dp_vs_service *svc, struct dp_vs_dest_conf *udest,
                                              struct dp_vs_dest **dest_p);

struct dp_vs_dest *dp_vs_lookup_dest(int af, struct dp_vs_service *svc,
                                     const union inet_addr *daddr, uint16_t dport);

int dp_vs_add_dest(struct dp_vs_service *svc, struct dp_vs_dest_conf *udest);

int dp_vs_edit_dest(struct dp_vs_service *svc, struct dp_vs_dest_conf *udest);

void dp_vs_unlink_dest(struct dp_vs_service *svc,
                        struct dp_vs_dest *dest, int svcupd);

void dp_vs_dest_put(struct dp_vs_dest *dest);

int dp_vs_del_dest(struct dp_vs_service *svc, struct dp_vs_dest_conf *udest);

int dp_vs_get_dest_entries(const struct dp_vs_service *svc,
                           struct dp_vs_get_dests *uptr);

int dp_vs_dest_init(void);

int dp_vs_dest_term(void);

#endif /* __DPVS_DEST_H__ */
